package KEGG;
use strict;
require Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw( setLibrary loadKOs loadKOGenes KEGGSort loadSpecies getSpeciesGenes getSpeciesKOs loadKOPathways getSpeciesPathways getAverageKOLengths getNormalizedGeneLengths );

my $Library = '/net/gs/vol3/shared/borensteinlab/DATA_RAW/KEGG/KEGG_2011_11_2';
my $ParsedLibrary = '/net/gs/vol3/shared/borensteinlab/DATA_PARSED/KEGG/KEGG_PARSED_2011_11_2';

# Changes the internal library
sub setLibrary {
  my $location = shift;
  
  unless ( defined $location ) {
    print "KEGG.pm - setLibrary: Must input the KEGG pathway.\n";
    return 0;
  }
  
  unless ( -d $location ) {
    print "KEGG.pm - setLibrary: KEGG pathway does not exist: \"$location\".\n";
    return 0;
  }
  
  $Library = $location;
  
  return 1;
}

# Based on Elhanan's code
# Run with:
#   my %geneToKO = %{ loadKOs($pathToDatabase) };

sub loadKOs {
  my %geneToKO;

  my $koFile = shift;

  unless ( defined $koFile ) {
#    print "KEGG.pm - loadKOs: Must input the KEGG KO Database\n";
#    return \%geneToKO;
    $koFile = $Library . '/ANNOTATIONS/KEGG_KO/ko';
  }
  
  unless ( -e $koFile ) {
    print "KEGG.pm - loadKOs: koFile \"$koFile\" does not exist.\n";
    return \{};
  }

  my $end_sign = "///";

  #=============================================================
  # Read and analyze the ko (BIG) file.
  # =============================================================
  open (FH, $koFile) or die "Cannot read $koFile\n";

  # create a hash with gene names as keys and KO numbers as values

  my $c_ko;
  while (<FH>) {
    my $line=$_; 
    chomp($line);
  
    if ($line =~ m/ENTRY /) {  
      ($c_ko) = ($line =~ m/ENTRY[ ]+(\S+)/); 
    }
  
    if ($line =~ m/GENES /) {  
    
      while ($line =~ m/GENES / or substr($line,0,12) eq "            ") { 
	my ($mega_line) = substr($line,12); 
	
	if (substr($mega_line,0,3) ne "   ") { 
	  my ($c_sp,$genelist) = $mega_line =~ m/(\S+): (.+)/; 
	  $c_sp = lc $c_sp;

	  # Remove the parentheses prior to split'ing
	  1 while $genelist =~ s/\([^(]+?\)//g;

	  my @genes = split(/ /,$genelist); # splits inside ()s
	  foreach my $gene (@genes) {
#	    if ($gene =~ m/(.+?)\(.+/) {
#	      $gene = $2;
#	    }
	    my $fullgene = $c_sp . ":" . $gene;	
	    $geneToKO{$fullgene} = $c_ko;	
	  }
	}  		   		   	
      
	$line = <FH>;
      }
    
    }
  
  }
  close (FH);

  # #Debug
  # print "Done parsing KOs\n";
  # print "Found ", scalar (keys %geneToKO), " KOs\n";
  # print "Testing\t", $geneToKO{lc('dana_gf14888')}, "\n";
  # print "Testing\t", $geneToKO{lc('Dana_GF14888(Dana_Adh)')}, "\n";
  # print "Testing\t", $geneToKO{lc('ATH:AT1G32780')}, "\n";

  # foreach ( keys %geneToKO ) {
  #   print $_, "\t", $geneToKO{$_}, "\n";
  # }

  return \%geneToKO;
}

sub loadKOGenes {
  my %genes;

  my $koFile = shift;
  
  unless ( defined $koFile ) {
    $koFile = $Library . '/ANNOTATIONS/KEGG_KO/ko';
  }

  unless ( -e $koFile ) {
    print "KEGG.pm - loadKOGenes: koFile \"$koFile\" does not exist.\n";
    return \{};
  }

  my $end_sign = "///";

  #=============================================================
  # Read and analyze the ko (BIG) file.
  # =============================================================
  open (FH, $koFile) or die "Cannot read $koFile\n";

  # create a hash with gene names as keys and KO numbers as values

  my $c_ko;
  while (<FH>) {
    my $line=$_; 
    chomp($line);
  
    if ($line =~ m/ENTRY /) {  
      ($c_ko) = ($line =~ m/ENTRY[ ]+(\S+)/); 
    }
  
    if ($line =~ m/GENES /) {  
#      my $c_sp = "";
    
      while ($line =~ m/GENES / or substr($line,0,12) eq "            ") { 
	my ($mega_line) = substr($line,12); 
      
	if (substr($mega_line,0,3) ne "   ") { 
	  my ($c_sp,$genelist) = $mega_line =~ m/(\S+): (.+)/; 
	  $c_sp = lc $c_sp;

	  # Remove the parentheses prior to split'ing
	  1 while $genelist =~ s/\([^(]*?\)//g;

	  my @genes = split(/ /,$genelist);
	  foreach my $gene (@genes) {
	    my $fullgene = $c_sp . ":" . $gene;	
	    push @{$genes{$c_ko}}, $fullgene;
	  }
	}	   	
	
	$line = <FH>;
      }
    
    }
  
  }
  close (FH);

  return \%genes;
}

# Load the mappings from KOs to Pathways
# There may be multiple pathways for a KO.
sub loadKOPathways {
  my %path;

  my $pathFile = $Library . '/ANNOTATIONS/KEGG_KO/ko_pathway.list';
  
  unless ( -e $pathFile ) {
    print "KEGG.pm - loadKOPathways: pathway file \"$pathFile\" does not exist.\n";
    return \{};
  }

  # --- Read through the pathway file --- #

  open (FH, $pathFile) or die "Cannot read $pathFile\n";

  # create a hash with gene names as keys and KO numbers as values

  while (<FH>) {
    chomp;

    next unless (/^ko/);
    
    #ko:K00004       path:ko00650
    my ($ko,$pathway) = split(/\s+/);
    
    unless (defined $ko) {
      print "KEGG.pm - loadKOPathways: Cannot read line ( $_ ).\n";
      next;
    }
    
    $ko =~ s/^ko://;
    $pathway =~ s/^path://;
    
    push @{$path{$ko}}, $pathway;
  }
  close (FH);

  return \%path;
}

# Use like my $kos_sorted_by_number = sort KEGGSort @list_of_kos
sub KEGGSort { 
  # We have to magic a and b to us; this should be faster that pulling them
  my ($pkg)=caller; 
  no strict 'refs';
  return (${$pkg.'::a'}=~/^[A-Za-z]+(\d+)/)[0] <=> (${$pkg.'::b'}=~/^[A-Za-z]+(\d+)/)[0];
}


sub loadSpecies {

  my %species;
  my %taxonomy;
  
  my $genomeFile = shift;
  
  unless ( defined $genomeFile ) {
    $genomeFile = $Library . '/GENOME_INFO/genome';
  }

  unless ( -e $genomeFile ) {
    print "KEGG.pm - loadSpecies: genomeFile \"$genomeFile\" does not exist.\n";
    return \{};
  }


  # --- Read through the Genome file --- #

  open (FH, $genomeFile) or die "Cannot read $genomeFile\n";


  while (<FH>) {
    chomp;
  
    if (/^ENTRY/) {  
      my ($entry,$short,$long,$definition,$lineage,$taxid,$shorter,$sequence);
      ($entry) = /^ENTRY\s+(\S+)/;
      my $line = $_;
      while ( $line !~ /^\/\/\// ) {
	$line = <FH>; chomp($line);

	if ($line =~ /^NAME/) {
	  ($short,$long,$shorter,$taxid) = $line =~ /^NAME\s+([^,]+), ([^,]+)(, [A-Za-z0-9]+)?, (\d+)\s*$/;
	} elsif ($line =~ /^DEFINITION/) {
	  ($definition) = $line =~ /DEFINITION\s+(.+)$/;
	} elsif ($line =~ /^\s{2}LINEAGE/) {
	  ($lineage) = $line =~ /^\s{2}LINEAGE\s+(.+)/;
	  
	  while ( $lineage =~ /;\s*$/ ) { # we need the next line
	    $line = <FH>; chomp($line);
	    $lineage .= $line;
	  }
	  
	  $lineage =~ s/;\s+/;/g; # Get ready for a split
	  
	}
      }

      unless ( defined $entry && defined $short && defined $long && defined $definition && defined $lineage && defined $taxid ) {
	print "KEGG.pm loadSpecies: serious problem reading species file! Exiting.\n";
	print "entry $entry\n";
	print "short $short long $long taxid $taxid\n";
	print "definition $definition\n";
	print "lineage $lineage\n";
	exit;
      }
      
      $taxonomy{$taxid}=$short;
      $species{$short}->{entry} = $entry;
      $species{$short}->{long} = lc($long);
      $species{$short}->{short} = $short; # For passing by itself
      $species{$short}->{definition} = $definition;
      $species{$short}->{lineage} = $lineage;
      $species{$short}->{domain} = (split(';',$lineage))[0];

      # -- Get genus and species
      my @name = split(/\s/,$definition);

      my $i=0; # This gets a bit weird
      $species{$short}->{genus} = $name[$i++];
      
      if ($name[$i-1] =~ /^Candidatus$/) { # ie. Candidatus Somethingicus
	$species{$short}->{genus} .= ' ' . $name[$i++];
      }
      if ( $name[$i] =~ /^\(/ ) { # Subgenus or other alternative lifestyle
	$i++;
      }
      
      $species{$short}->{species} = $name[$i++];
      
      if ( $name[$i-1] =~ /^sp\.$/i || $name[$i-1] =~ /^species$/i ) {
	while ( $i < scalar @name && $name[$i] !~ /^subsp/i && $name[$i] !~ /^\(/i ) {
	  $species{$short}->{species} .= ' ' . $name[$i++];
	}
      } elsif ($name[$i-1] =~ /^genomosp/i) {
	$species{$short}->{species} =~ s/genomospecies/genomesp\./;
	$species{$short}->{species} .= ' ' . $name[$i++];
      }
      
      # Clean up
      $species{$short}->{species} =~ s/,//;

    }
  }
  close (FH);

  return (\%species, \%taxonomy);
}

sub getSpeciesGenes {

  my $species = shift;
  
  my %genes;

  unless ( defined $species && defined $species->{long} && defined $species->{short} ) {
    print "KEGG.pm usage: getSpeciesGenes species-hash\n";
    return \%genes;
  }
  
  my $speciesFile = $Library . '/ORGANISMS/' . $species->{short} . '/' . $species->{long} . '.pos';

  unless ( -e $speciesFile ) {
    print "KEGG.pm - getSpeciesGenes: Cannot locate genome species file $speciesFile.\n";
    return \%genes;
  }
  
  open (FH, $speciesFile) or die "Cannot read $speciesFile\n";
  while (<FH>) {
    chomp;
    
    # gene[0]  name[1] beg[2]     span[3]      length[4]
    #BU001	gidA	197	197..2083	1887	
    my ($gene,$length) = (split /\t/)[0,4];
    
    next unless (defined $gene && defined $length);
    
    $gene = $species->{short} . ':' . $gene;
    
    $genes{$gene}->{length} = $length;
    
  }
  
  return \%genes;
}

sub getSpeciesKOs {

  my $species = shift;
  
  my %kos;

  unless ( defined $species && defined $species->{short} ) {
    print "KEGG.pm usage: getSpeciesKOs species-hash\n";
    return \%kos;
  }
  
  my $koFile = $Library . '/ORGANISMS/' . $species->{short} . '/' . $species->{short} . '_ko.list';

  unless ( -e $koFile ) {
    print "KEGG.pm - getSpeciesGenes: Cannot locate genome species file $koFile.\n";
    return \%kos;
  }
  
  open (FH, $koFile) or die "Cannot read $koFile\n";
  while (<FH>) {
    chomp;
    
    #cte:CT0001	ko:K02338
    my ($gene,$ko) = split /\t/;
    #$gene =~ s/($species->{short})://i;
    $ko =~ s/ko://i;
    
    next unless (defined $gene && defined $ko);
    
    $kos{$gene} = $ko;
    
  }
  
  return \%kos;
}

# Note that genes can have multiple Pathways, while they *should* only have one KO.
sub getSpeciesPathways {

  my $species = shift;
  
  my %pathways;

  unless ( defined $species && defined $species->{short} ) {
    print "KEGG.pm usage: getSpeciesPathways species-hash\n";
    return \%pathways;
  }
  
  my $koFile = $Library . '/ORGANISMS/' . $species->{short} . '/' . $species->{short} . '_pathway.list';

  unless ( -e $koFile ) {
    print "KEGG.pm - getSpeciesGenes: Cannot locate genome species file $koFile.\n";
    return \%pathways;
  }
  
  open (FH, $koFile) or die "Cannot read $koFile\n";
  while (<FH>) {
    chomp;
    
    #buc:CT0001	path:K02338
    my ($gene,$path) = split /\t/;
    #$gene =~ s/($species->{short})://i;
    $path =~ s/path:[A-Za-z]+//i;
    $path = 'ko' . $path;
    
    next unless (defined $gene && defined $path);
    
    push @{$pathways{$gene}}, $path;
    
  }
  
  return \%pathways;
}

sub getAverageKOLengths {
  
  my $which = shift;
  unless (defined $which) {
    $which='';
  }
  
  my %lengths;

  my $KODistributionFile = $ParsedLibrary . '/KO_Distributions.txt' . $which . '.gz';
  
  unless ( -e $KODistributionFile ) {
    print "KEGG.pm, getAverageKOLength: - Cannot locate $KODistributionFile\n";
    return \%lengths;
  }
  
  open (FH, "zcat $KODistributionFile | ") or die "Cannot read $KODistributionFile\n";
  while (<FH>) {
    chomp;
    
    next if ( /^#/ );

    my ($ko,$avg) = (split /\t/)[0,1] ;
    
    unless (defined $ko && defined $avg) {
      print "KEGG.pm, getAverageKOLength: Bad line: $_\n";
      next;
    }
    
#    print "ko($ko) = avg($avg)\n";
    
    $lengths{$ko} = $avg;
    
  }  
  return \%lengths;
}

sub getNormalizedGeneLengths {
  
  my %lengths;

  my $GeneLengthFile = $ParsedLibrary . '/Gene_Lengths.txt.gz';
  
  unless ( -e $GeneLengthFile ) {
    print "KEGG.pm, getAverageKOLength: - Cannot locate $GeneLengthFile\n";
    return \%lengths;
  }
  
  open (FH, "zcat $GeneLengthFile | ") or die "Cannot read $GeneLengthFile\n";
  while (<FH>) {
    chomp;
    
    next if ( /^#/ );

    my ($gene,$avg) = (split /\t/)[0,2] ;
    
    unless (defined $gene && defined $avg) {
      print "KEGG.pm, getAverageKOLength: Bad line: $_\n";
      next;
    }
    
#    print "ko($gene) = avg($avg)\n";
    
    $lengths{$gene} = $avg;
    
  }  
  return \%lengths;
}


1;
